@@ -6,7 +6,7 @@
 #ifndef _IP_VS_H
 #define _IP_VS_H
 
-#define IP_VS_VERSION_CODE            0x000102
+#define IP_VS_VERSION_CODE            0x000200
 #define NVERSION(version)                       \
         (version >> 16) & 0xFF,                 \
         (version >> 8) & 0xFF,                  \
@@ -93,7 +93,7 @@ struct ip_vs_rule_user {
 #define IP_VS_CONN_F_INACTIVE         0x0100    /* not established */
 #define IP_VS_CONN_F_OUT_SEQ          0x0200    /* must do output seq adjust */
 #define IP_VS_CONN_F_IN_SEQ           0x0400    /* must do input seq adjust */
-#define IP_VS_CONN_F_NO_CPORT         0x0800    /* no cport set yet */
+#define IP_VS_CONN_F_NO_CPORT         0x0800    /* no client port set yet */
 
 
 #ifdef __KERNEL__
@@ -178,6 +178,7 @@ enum {
 	NET_IPV4_VS_LBLC_EXPIRE=19,
 	NET_IPV4_VS_LBLCR_EXPIRE=20,
 	NET_IPV4_VS_USE_DST=21,
+	NET_IPV4_VS_CACHE_BYPASS=22,
 };
 
 enum {
@@ -248,6 +249,10 @@ struct ip_vs_conn {
 	struct ip_vs_timeout_table *timeout_table;
 	struct ip_vs_dest       *dest;          /* real server */
         atomic_t                in_pkts;        /* incoming packet counter */
+
+        /* packet transmitter for different forwarding methods */
+        int (*packet_xmit)(struct sk_buff *skb, struct ip_vs_conn *cp);
+
         /* Note: we can group the following members into a structure,
            in order to save more space, and the following members are
            only used in VS/NAT anyway */
@@ -360,7 +365,8 @@ struct ip_vs_app
  */
 extern struct ip_vs_conn * ip_vs_schedule(struct ip_vs_service *svc,
                                           struct iphdr *iph);
-const char *vs_proto_name(unsigned proto);
+extern int ip_vs_proto_doff(unsigned proto, char *th, unsigned size);
+extern const char *ip_vs_proto_name(unsigned proto);
 
 
 /*
@@ -397,12 +403,12 @@ extern struct ip_vs_conn *
 ip_vs_conn_new(int proto, __u32 maddr, __u16 mport, __u32 saddr,
                __u16 sport, __u32 daddr, __u16 dport, unsigned flags);
 extern const char * ip_vs_state_name(int state);
-extern int vs_set_state(struct ip_vs_conn *cp, int state_off,
-                        struct iphdr *iph, void *tp);
+extern int ip_vs_set_state(struct ip_vs_conn *cp, int state_off,
+                           struct iphdr *iph, void *tp);
 extern int ip_vs_conn_listen(struct ip_vs_conn *cp);
 extern int ip_vs_check_template(struct ip_vs_conn *ct);
 extern void ip_vs_secure_tcp_set(int on);
-extern void ip_vs_set_state(struct ip_vs_conn *cp, int new_state);
+extern void ip_vs_bind_bypass_xmit(struct ip_vs_conn *cp);
 extern void ip_vs_bind_dest(struct ip_vs_conn *cp, struct ip_vs_dest *dest);
 extern void ip_vs_unbind_dest(struct ip_vs_conn *cp);
 extern void ip_vs_random_dropentry(void);
@@ -425,7 +431,7 @@ static inline void ip_vs_control_del(struct ip_vs_conn *cp)
                 return;
 	}
         
-	IP_VS_DBG(1, "DELeting control for: "
+	IP_VS_DBG(7, "DELeting control for: "
                   "cp.dst=%d.%d.%d.%d:%d ctl_cp.dst=%d.%d.%d.%d:%d\n",
                   NIPQUAD(cp->caddr),ntohs(cp->cport),
                   NIPQUAD(ctl_cp->caddr),ntohs(ctl_cp->cport));
@@ -452,7 +458,7 @@ ip_vs_control_add(struct ip_vs_conn *cp, struct ip_vs_conn *ctl_cp)
 		ip_vs_control_del(cp);
 	}
         
-	IP_VS_DBG(1, "ADDing control for: "
+	IP_VS_DBG(7, "ADDing control for: "
                   "cp.dst=%d.%d.%d.%d:%d ctl_cp.dst=%d.%d.%d.%d:%d\n",
                   NIPQUAD(cp->caddr),ntohs(cp->cport),
                   NIPQUAD(ctl_cp->caddr),ntohs(ctl_cp->cport));
@@ -500,12 +506,17 @@ extern struct ip_vs_scheduler *ip_vs_lookup_scheduler(const char *sched_name);
  *      IPVS control data and functions
  *      (from ip_vs_ctl.c)
  */
-extern rwlock_t __ip_vs_svc_lock;
+extern int sysctl_use_dst;
+extern int sysctl_ip_vs_cache_bypass;
 extern atomic_t ip_vs_dropentry;
-extern struct ip_vs_service *
-ip_vs_lookup_service(unsigned long fwmark,
-                     __u16 protocol, __u32 vaddr, __u16 vport);
-extern struct ip_vs_service * ip_vs_lookup_svc_fwm(unsigned long fwmark);
+extern struct ip_vs_service *ip_vs_service_get(unsigned long fwmark,
+                                               __u16 protocol,
+                                               __u32 vaddr, __u16 vport);
+static inline void ip_vs_service_put(struct ip_vs_service *svc)
+{
+        atomic_dec(&svc->refcnt);
+}
+
 extern struct ip_vs_dest *
 ip_vs_lookup_real_service(__u16 protocol, __u32 daddr, __u16 dport);
 extern void update_defense_level(void);
