@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_conn.c,v 1.7 2000/12/19 15:04:23 wensong Exp $
+ * Version:     $Id: ip_vs_conn.c,v 1.9 2001/01/10 07:42:53 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -35,6 +35,13 @@
 #include <linux/in.h>
 #include <linux/proc_fs.h>              /* for proc_net_* */
 #include <asm/softirq.h>                /* for local_bh_* */
+#include <net/ip.h>
+#include <net/tcp.h>                    /* for csum_tcpudp_magic */
+#include <net/udp.h>
+#include <net/icmp.h>                   /* for icmp_send */
+#include <net/route.h>                  /* for ip_route_output */
+#include <linux/netfilter.h>
+#include <linux/netfilter_ipv4.h>
 
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
@@ -240,8 +247,8 @@ struct ip_vs_conn *ip_vs_conn_in_get
         if (!ret && atomic_read(&ip_vs_conn_no_cport_cnt))
                 ret = __ip_vs_conn_in_get(protocol, s_addr, 0, d_addr, d_port);
         
-        IP_VS_DBG(2, "lookup/in %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
-                  vs_proto_name(protocol),
+        IP_VS_DBG(7, "lookup/in %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
+                  ip_vs_proto_name(protocol),
                   NIPQUAD(s_addr), ntohs(s_port),
                   NIPQUAD(d_addr), ntohs(d_port),
                   ret?"hit":"not hit");
@@ -287,8 +294,8 @@ struct ip_vs_conn *ip_vs_conn_out_get
 
         ct_read_unlock(hash);
         
-        IP_VS_DBG(2, "lookup/out %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
-                  vs_proto_name(protocol),
+        IP_VS_DBG(7, "lookup/out %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d %s\n",
+                  ip_vs_proto_name(protocol),
                   NIPQUAD(s_addr), ntohs(s_port),
                   NIPQUAD(d_addr), ntohs(d_port),
                   ret?"hit":"not hit");
@@ -531,7 +538,7 @@ vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
 	} 
 
 	if ((state_idx = vs_tcp_state_idx(th, state_off)) < 0) {
-		IP_VS_DBG(1, "vs_tcp_state_idx(%d)=%d!!!\n",
+		IP_VS_DBG(8, "vs_tcp_state_idx(%d)=%d!!!\n",
                           state_off, state_idx);
 		goto tcp_state_out;
 	}
@@ -542,9 +549,9 @@ vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
 	if (new_state != cp->state) {
                 struct ip_vs_dest *dest = cp->dest;
 
-                IP_VS_DBG(1, "%s %s [%c%c%c%c] %u.%u.%u.%u:%d->"
+                IP_VS_DBG(8, "%s %s [%c%c%c%c] %u.%u.%u.%u:%d->"
                           "%u.%u.%u.%u:%d state: %s->%s cnt:%d\n",
-                          vs_proto_name(cp->protocol),
+                          ip_vs_proto_name(cp->protocol),
                           (state_off==VS_STATE_OUTPUT)?"output ":"input ",
                           th->syn? 'S' : '.',
                           th->fin? 'F' : '.',
@@ -577,8 +584,8 @@ vs_tcp_state(struct ip_vs_conn *cp, int state_off, struct tcphdr *th)
 /*
  *	Handle state transitions
  */
-int vs_set_state(struct ip_vs_conn *cp,
-                 int state_off, struct iphdr *iph, void *tp)
+int ip_vs_set_state(struct ip_vs_conn *cp,
+                    int state_off, struct iphdr *iph, void *tp)
 {
 	switch (iph->protocol) {
 		case IPPROTO_ICMP:
@@ -602,23 +609,483 @@ int ip_vs_conn_listen(struct ip_vs_conn *cp)
 }
 
 
+/*
+ *      Bypass transmitter
+ *      Let packets bypass the destination when the destination is not
+ *      available, it may be only used in transparent cache cluster.
+ */
+static int ip_vs_bypass_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
+{
+	struct rtable *rt;			/* Route to the other host */
+	struct iphdr  *iph = skb->nh.iph;
+	u8     tos = iph->tos;
+        int    mtu;
+
+	EnterFunction(10);
+
+        if (ip_route_output(&rt, iph->daddr, 0, RT_TOS(tos), 0)) {
+                IP_VS_DBG(0, "ip_vs_bypass_xmit(): route error, "
+                          "dest: %u.%u.%u.%u\n", NIPQUAD(iph->daddr));
+                goto tx_error_icmp;
+        }
+
+        /* MTU checking */
+        mtu = rt->u.dst.pmtu;
+        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+                ip_rt_put(rt);
+                IP_VS_DBG(0, "ip_vs_bypass_xmit(): frag needed\n");
+                goto tx_error;
+        }
+
+	/* drop old route */
+        dst_release(skb->dst);
+	skb->dst = &rt->u.dst;
+
+#ifdef CONFIG_NETFILTER_DEBUG
+        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
+#endif /* CONFIG_NETFILTER_DEBUG */
+        ip_send(skb);
+
+        LeaveFunction(10);
+	return NF_STOLEN;
+
+  tx_error_icmp:
+	dst_link_failure(skb);
+  tx_error:
+	kfree_skb(skb);
+	return NF_STOLEN;
+}
+
+
+/*
+ *  Bind a connection entry with the passby_xmit.
+ *  Called by ip_vs_leave.
+ */
+void ip_vs_bind_bypass_xmit(struct ip_vs_conn *cp)
+{
+        cp->packet_xmit = ip_vs_bypass_xmit;
+}
+
+
+/*
+ *      NULL transmitter (do nothing except return NF_ACCEPT)
+ */
+static int ip_vs_null_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
+{
+        return NF_ACCEPT;
+}
+
+
+/*
+ *      NAT transmitter (only for outside-to-inside nat forwarding)
+ */
+static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
+{
+        struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
+	struct iphdr  *iph = skb->nh.iph;
+	u8     tos = iph->tos;
+	u32    rtos;
+	union ip_vs_tphdr h;
+	unsigned short size;
+	int doff;
+	int csum = 0;
+	int csum_ok = 0;
+        int mtu;
+
+	EnterFunction(10);
+
+        h.raw = (char*) iph + iph->ihl * 4;
+	size = ntohs(iph->tot_len) - (iph->ihl * 4);
+	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
+
+        /*
+         *  Must do TCP/UDP checksum checking before mangling the packet
+         */
+        if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
+                switch (skb->ip_summed) {
+                case CHECKSUM_NONE:
+                        csum = csum_partial(h.raw + doff, size - doff, 0);
+                        csum_ok++;
+                        skb->csum = csum_partial(h.raw , doff, csum);
+
+                case CHECKSUM_HW:
+                        if (csum_tcpudp_magic(iph->saddr, iph->daddr,
+                                              size, iph->protocol, skb->csum))
+                        {
+                                IP_VS_DBG(0, "Incoming failed %s checksum "
+                                          "from %d.%d.%d.%d (size=%d)!\n",
+                                          ip_vs_proto_name(iph->protocol),
+                                          NIPQUAD(iph->saddr),
+                                          size);
+                                goto tx_error;
+                        }
+                default:/* CHECKSUM_UNNECESSARY */
+                }
+        }
+        
+        /*
+         *  Check if it is no_cport connection (VS/NAT only)...
+         */
+        if (cp->flags & IP_VS_CONN_F_NO_CPORT ) {
+                atomic_dec(&ip_vs_conn_no_cport_cnt);
+                ip_vs_conn_unhash(cp);
+                cp->flags &= ~IP_VS_CONN_F_NO_CPORT;
+                cp->cport = h.portp[0];
+                /* hash on new dport */
+                ip_vs_conn_hash(cp);
+
+                IP_VS_DBG(10, "filled cport=%d\n", ntohs(cp->dport));
+        }
+                
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_nat_xmit(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(10, "NAT: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+                          NIPQUAD(dest->addr),
+                          atomic_read(&rt->u.dst.__refcnt), rtos);
+	}
+	spin_unlock(&dest->dst_lock);
+
+        /* MTU checking */
+        mtu = rt->u.dst.pmtu;
+        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+                ip_rt_put(rt);
+                IP_VS_DBG(0, "ip_vs_nat_xmit(): frag needed\n");
+                goto tx_error;
+        }
+        
+	/* drop old route */
+        dst_release(skb->dst);
+	skb->dst = &rt->u.dst;
+
+        /* mangle the packet */
+        iph->daddr = cp->daddr;
+        h.portp[1] = cp->dport;
+
+        /*
+         *	Invalidate csum saving if it has ip_vs_app helper
+         */
+        if (cp->app)
+                csum_ok = 0;
+
+        /*
+         *	Attempt ip_vs_app call.
+         *	will fix ip_vs_conn and iph ack_seq stuff
+         */
+        if (ip_vs_app_pkt_in(cp, &skb) != 0) {
+                /*
+                 *	skb has changed, update pointers.
+                 */
+                /*  skb = *skb_p; */
+                iph = skb->nh.iph;
+                h.raw = (char*) iph + iph->ihl*4;
+                size = ntohs(iph->tot_len) - (iph->ihl * 4);
+        }
+
+        /*
+         * Yug! adjust UDP/TCP checksums
+         */
+        /*
+         *	Transport's payload partial csum
+         */
+        if (!csum_ok) {
+                csum = csum_partial(h.raw + doff, size - doff, 0);
+        }
+        skb->csum = csum;
+
+        /*
+         *	Protocol csum
+         */
+        switch (iph->protocol) {
+        case IPPROTO_TCP:
+                h.th->check = 0;
+                h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
+                                              size, iph->protocol,
+                                              csum_partial(h.raw, doff, csum));
+                break;
+        case IPPROTO_UDP:
+                h.uh->check = 0;
+                h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
+                                              size, iph->protocol,
+                                              csum_partial(h.raw, doff, csum));
+                if (h.uh->check == 0)
+                        h.uh->check = 0xFFFF;
+                break;
+        }
+        ip_send_check(iph);
+
+        IP_VS_DBG(10, "NAT to %u.%u.%u.%u:%d\n",
+                  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
+
+        /* FIXME: when application helper enlarges the packet and the length
+           is larger than the MTU of outgoing device, there will be still
+           MTU problem. */
+
+#ifdef CONFIG_NETFILTER_DEBUG
+        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
+#endif /* CONFIG_NETFILTER_DEBUG */
+        ip_send(skb);
+
+        LeaveFunction(10);
+	return NF_STOLEN;
+
+  tx_error_icmp:
+	dst_link_failure(skb);
+  tx_error:
+	kfree_skb(skb);
+	return NF_STOLEN;
+}
+
+
+/*
+ *   IP Tunneling transmitter
+ *
+ *   This function encapsulates the packet in a new IP packet, its
+ *   destination will be set to cp->daddr. Most code of this function
+ *   is taken from ipip.c.
+ *
+ *   It is used in VS/TUN cluster. The load balancer selects a real
+ *   server from a cluster based on a scheduling algorithm,
+ *   encapsulates the request packet and forwards it to the selected
+ *   server. For example, all real servers are configured with
+ *   "ifconfig tunl0 <Virtual IP Address> up". When the server receives
+ *   the encapsulated packet, it will decapsulate the packet, processe
+ *   the request and return the response packets directly to the client
+ *   without passing the load balancer. This can greatly increase the
+ *   scalability of virtual server.
+ */
+static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
+{
+	struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
+	struct net_device *tdev;		/* Device to other host */
+	struct iphdr  *old_iph = skb->nh.iph;
+	u8     tos = old_iph->tos;
+	u32    rtos;
+	u16    df = old_iph->frag_off;
+	struct iphdr  *iph;			/* Our new IP header */
+	int    max_headroom;			/* The extra header space needed */
+	int    mtu;
+
+	EnterFunction(10);
+
+        if (skb->protocol != __constant_htons(ETH_P_IP)) {
+		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): protocol error, "
+                          "ETH_P_IP: %d, skb protocol: %d\n",
+			  __constant_htons(ETH_P_IP), skb->protocol);
+		goto tx_error;
+	}
+
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(10, "TUN: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+                          NIPQUAD(dest->addr),
+                          atomic_read(&rt->u.dst.__refcnt), rtos);
+	}
+	spin_unlock(&dest->dst_lock);
+
+	tdev = rt->u.dst.dev;
+
+	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
+	if (mtu < 68) {
+		ip_rt_put(rt);
+		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): mtu less than 68\n");
+		goto tx_error;
+	}
+	if (skb->dst && mtu < skb->dst->pmtu)
+		skb->dst->pmtu = mtu;
+
+	df |= (old_iph->frag_off&__constant_htons(IP_DF));
+
+	if ((old_iph->frag_off&__constant_htons(IP_DF))
+            && mtu < ntohs(old_iph->tot_len)) {
+		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+		ip_rt_put(rt);
+		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): frag needed\n");
+		goto tx_error;
+	}
+
+	skb->h.raw = skb->nh.raw;
+
+	/*
+	 * Okay, now see if we can stuff it in the buffer as-is.
+	 */
+	max_headroom = (((tdev->hard_header_len+15)&~15)+sizeof(struct iphdr));
+
+	if (skb_headroom(skb) < max_headroom
+            || skb_cloned(skb) || skb_shared(skb)) {
+		struct sk_buff *new_skb =
+                        skb_realloc_headroom(skb, max_headroom);
+		if (!new_skb) {
+			ip_rt_put(rt);
+			kfree_skb(skb);
+			IP_VS_ERR("ip_vs_tunnel_xmit(): no memory\n");
+			return -EINVAL;
+		}
+		kfree_skb(skb);
+		skb = new_skb;
+	}
+
+	skb->nh.raw = skb_push(skb, sizeof(struct iphdr));
+	memset(&(IPCB(skb)->opt), 0, sizeof(IPCB(skb)->opt));
+
+	/* drop old route */
+	dst_release(skb->dst);
+	skb->dst = &rt->u.dst;
+
+	/*
+	 *	Push down and install the IPIP header.
+	 */
+	iph			=	skb->nh.iph;
+	iph->version		=	4;
+	iph->ihl		=	sizeof(struct iphdr)>>2;
+	iph->frag_off		=	df;
+	iph->protocol		=	IPPROTO_IPIP;
+	iph->tos		=	tos;
+	iph->daddr		=	rt->rt_dst;
+	iph->saddr		=	rt->rt_src;
+	iph->ttl		=	old_iph->ttl;
+	iph->tot_len		=	htons(skb->len);
+	ip_select_ident(iph, &rt->u.dst);
+	ip_send_check(iph);
+
+#ifdef CONFIG_NETFILTER_DEBUG
+        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
+#endif /* CONFIG_NETFILTER_DEBUG */
+	ip_send(skb);
+
+        LeaveFunction(10);
+        
+	return NF_STOLEN;
+
+  tx_error_icmp:
+	dst_link_failure(skb);
+  tx_error:
+	kfree_skb(skb);
+	return NF_STOLEN;
+}
+
+
+/*
+ *      Direct Routing transmitter
+ */
+static int ip_vs_dr_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
+{
+	struct rtable *rt;			/* Route to the other host */
+	struct ip_vs_dest *dest;
+	struct iphdr  *iph = skb->nh.iph;
+	u8     tos = iph->tos;
+	u32    rtos;
+        int    mtu;
+
+	EnterFunction(10);
+
+	dest = cp->dest;
+	rtos = RT_TOS(tos);
+	spin_lock(&dest->dst_lock);
+	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
+		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
+			spin_unlock(&dest->dst_lock);
+			IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, "
+				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
+			goto tx_error_icmp;
+		}
+		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
+		IP_VS_DBG(10, "DR: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+                          NIPQUAD(dest->addr),
+                          atomic_read(&rt->u.dst.__refcnt), rtos);
+	}
+	spin_unlock(&dest->dst_lock);
+
+        /* MTU checking */
+        mtu = rt->u.dst.pmtu;
+        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
+                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
+                ip_rt_put(rt);
+                IP_VS_DBG(0, "ip_vs_dr_xmit(): frag needed\n");
+                goto tx_error;
+        }
+
+	/* drop old route */
+        dst_release(skb->dst);
+	skb->dst = &rt->u.dst;
+
+#ifdef CONFIG_NETFILTER_DEBUG
+        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
+#endif /* CONFIG_NETFILTER_DEBUG */
+        ip_send(skb);
+
+#if 0000
+        NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, rt->u.dst.dev,
+                do_ip_send);
+#endif
+        LeaveFunction(10);
+	return NF_STOLEN;
+
+  tx_error_icmp:
+	dst_link_failure(skb);
+  tx_error:
+	kfree_skb(skb);
+	return NF_STOLEN;
+}
+
+
 /*
  *  Bind a connection entry with a virtual service destination
- *  Called when a new connection entry is created for VS.
+ *  Called just after a new connection entry is created.
  */
 void ip_vs_bind_dest(struct ip_vs_conn *cp, struct ip_vs_dest *dest)
 {
-        cp->flags |= atomic_read(&dest->conn_flags);
-        cp->dest = dest;
-
         /*
          *    Increase the refcnt counter of the dest.
          */
         atomic_inc(&dest->refcnt);
 
-        IP_VS_DBG(1, "Bind-dest %s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
+        /* Bind with the destination and its corresponding transmitter */
+        cp->flags |= atomic_read(&dest->conn_flags);
+        cp->dest = dest;
+
+        switch (IP_VS_FWD_METHOD(cp)) {
+        case IP_VS_CONN_F_MASQ:
+                cp->packet_xmit = ip_vs_nat_xmit;
+                break;
+                
+        case IP_VS_CONN_F_TUNNEL:
+                cp->packet_xmit = ip_vs_tunnel_xmit;
+                break;
+                
+        case IP_VS_CONN_F_DROUTE:
+                cp->packet_xmit = ip_vs_dr_xmit;
+                break;
+                
+        case IP_VS_CONN_F_LOCALNODE:
+                cp->packet_xmit = ip_vs_null_xmit;
+                break;
+        }
+
+        IP_VS_DBG(9, "Bind-dest %s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
                   "d:%u.%u.%u.%u:%d fwd:%c s:%s flg:%X cnt:%d destcnt:%d\n",
-                  vs_proto_name(cp->protocol),
+                  ip_vs_proto_name(cp->protocol),
                   NIPQUAD(cp->caddr), ntohs(cp->cport),
                   NIPQUAD(cp->vaddr), ntohs(cp->vport),
                   NIPQUAD(cp->daddr), ntohs(cp->dport),
@@ -636,17 +1103,18 @@ void ip_vs_unbind_dest(struct ip_vs_conn *cp)
 {
         struct ip_vs_dest *dest = cp->dest;
         
-        IP_VS_DBG(1, "Unbind-dest %s c:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
-                  "d:%u.%u.%u.%u:%d fwd:%c s:%s flg:%X cnt:%d destcnt:%d\n",
-                  vs_proto_name(cp->protocol),
-                  NIPQUAD(cp->caddr), ntohs(cp->cport),
-                  NIPQUAD(cp->vaddr), ntohs(cp->vport),
-                  NIPQUAD(cp->daddr), ntohs(cp->dport),
-                  ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
-                  cp->flags, atomic_read(&cp->refcnt),
-                  atomic_read(&dest->refcnt));
-
         if (dest) {
+                IP_VS_DBG(9, "Unbind-dest %s c:%u.%u.%u.%u:%d "
+                          "v:%u.%u.%u.%u:%d d:%u.%u.%u.%u:%d fwd:%c "
+                          "s:%s flg:%X cnt:%d destcnt:%d",
+                          ip_vs_proto_name(cp->protocol),
+                          NIPQUAD(cp->caddr), ntohs(cp->cport),
+                          NIPQUAD(cp->vaddr), ntohs(cp->vport),
+                          NIPQUAD(cp->daddr), ntohs(cp->dport),
+                          ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
+                          cp->flags, atomic_read(&cp->refcnt),
+                          atomic_read(&dest->refcnt));
+
                 /*
                  * Decrease the inactconns or activeconns counter
                  * if it is not a connection template (cp->cport!=0).
@@ -683,10 +1151,10 @@ int ip_vs_check_template(struct ip_vs_conn *ct)
          */
         if ((dest == NULL) ||
             !(dest->flags & IP_VS_DEST_F_AVAILABLE)) {
-                IP_VS_DBG(1, "check_template: dest not available for "
+                IP_VS_DBG(9, "check_template: dest not available for "
                           "protocol %s s:%u.%u.%u.%u:%d v:%u.%u.%u.%u:%d "
                           "-> d:%u.%u.%u.%u:%d\n",
-                          vs_proto_name(ct->protocol),
+                          ip_vs_proto_name(ct->protocol),
                           NIPQUAD(ct->caddr), ntohs(ct->cport),
                           NIPQUAD(ct->vaddr), ntohs(ct->vport),
                           NIPQUAD(ct->daddr), ntohs(ct->dport));
@@ -774,7 +1242,7 @@ static void ip_vs_conn_expire(unsigned long data)
 	}
 
   expire_later:
-	IP_VS_DBG(1, "delayed: refcnt-1=%d conn.n_control=%d\n",
+	IP_VS_DBG(7, "delayed: refcnt-1=%d conn.n_control=%d\n",
                   atomic_read(&cp->refcnt)-1,
                   atomic_read(&cp->n_control));
 
@@ -872,7 +1340,7 @@ ip_vs_conn_getinfo(char *buffer, char **start, off_t offset, int length)
                         size = sprintf(buffer+len,
                                        "%-3s %08X %04X %08X %04X "
                                        "%08X %04X %-11s %7lu\n",
-                                       vs_proto_name(cp->protocol),
+                                       ip_vs_proto_name(cp->protocol),
                                        ntohl(cp->caddr), ntohs(cp->cport),
                                        ntohl(cp->vaddr), ntohs(cp->vport),
                                        ntohl(cp->daddr), ntohs(cp->dport),
@@ -974,13 +1442,13 @@ void ip_vs_random_dropentry(void)
 			 * Drop the entry, and drop its ct if not referenced
 			 */
 			ct_write_unlock(hash);
-			IP_VS_DBG(1, "Drop connection\n");
+			IP_VS_DBG(4, "Drop connection\n");
 			ct = cp->control;
 			fn = (cp->timer).function;
 			if (!del_sltimer(&cp->timer))
 				fn((unsigned long)cp);
 			if (ct && !atomic_read(&ct->n_control)) {
-				IP_VS_DBG(1, "Drop connection template\n");
+				IP_VS_DBG(4, "Drop connection template\n");
 				del_sltimer(&ct->timer);
 				fn((unsigned long)ct);
 			}
@@ -1004,7 +1472,7 @@ int ip_vs_conn_init(void)
         }
 
         /* Allocate ip_vs_conn slab cache */
-        ip_vs_conn_cachep = kmem_cache_create("ip_vs",
+        ip_vs_conn_cachep = kmem_cache_create("ip_vs_conn",
                                               sizeof(struct ip_vs_conn), 0,
                                               SLAB_HWCACHE_ALIGN, NULL, NULL);
         if (!ip_vs_conn_cachep) {
