@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_core.c,v 1.9 2000/12/19 15:04:23 wensong Exp $
+ * Version:     $Id: ip_vs_core.c,v 1.11 2001/01/10 07:42:53 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -49,14 +49,13 @@
 #include "ip_vs.h"
 //#include <net/ip_vs.h>
 
-extern int sysctl_use_dst;
 
 EXPORT_SYMBOL(register_ip_vs_scheduler);
 EXPORT_SYMBOL(unregister_ip_vs_scheduler);
 EXPORT_SYMBOL(ip_vs_bind_dest);
 EXPORT_SYMBOL(ip_vs_unbind_dest);
 EXPORT_SYMBOL(ip_vs_skb_replace);
-EXPORT_SYMBOL(vs_proto_name);
+EXPORT_SYMBOL(ip_vs_proto_name);
 EXPORT_SYMBOL(ip_vs_conn_new);
 EXPORT_SYMBOL(ip_vs_conn_in_get);
 EXPORT_SYMBOL(ip_vs_conn_out_get);
@@ -70,13 +69,14 @@ EXPORT_SYMBOL(ip_vs_get_debug_level);
 /* ID used in ICMP lookups */
 #define icmp_id(icmph)          ((icmph->un).echo.id)
 
+
 /*
  *	Get transport protocol data offset, check against size
  *	return:
  *		0  if other IP proto
  *		-1 if error
  */
-static inline int proto_doff(unsigned proto, char *th, unsigned size)
+int ip_vs_proto_doff(unsigned proto, char *th, unsigned size)
 {
 	int ret = -1;
 	switch (proto) {
@@ -112,7 +112,7 @@ static inline int proto_doff(unsigned proto, char *th, unsigned size)
 }
 
 
-const char *vs_proto_name(unsigned proto)
+const char *ip_vs_proto_name(unsigned proto)
 {
 	static char buf[20];
 
@@ -178,7 +178,7 @@ ip_vs_sched_persist(struct ip_vs_service *svc, struct iphdr *iph)
 	/* Mask saddr with the netmask to adjust template granularity */
 	snet = iph->saddr & svc->netmask;
 
-	IP_VS_DBG(1, "P-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u "
+	IP_VS_DBG(6, "P-schedule: src %u.%u.%u.%u:%u dest %u.%u.%u.%u:%u "
                   "mnet %u.%u.%u.%u\n",
 		  NIPQUAD(iph->saddr), ntohs(portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(portp[1]),
@@ -416,7 +416,7 @@ struct ip_vs_conn *ip_vs_schedule(struct ip_vs_service *svc, struct iphdr *iph)
 	 */
 	atomic_inc(&dest->inactconns);
 
-	IP_VS_DBG(1, "Schedule fwd:%c s:%s c:%u.%u.%u.%u:%u v:%u.%u.%u.%u:%u "
+	IP_VS_DBG(6, "Schedule fwd:%c s:%s c:%u.%u.%u.%u:%u v:%u.%u.%u.%u:%u "
 		  "d:%u.%u.%u.%u:%u flg:%X cnt:%d\n",
 		  ip_vs_fwd_tag(cp), ip_vs_state_name(cp->state),
 		  NIPQUAD(cp->caddr), ntohs(cp->cport),
@@ -438,17 +438,53 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 	struct iphdr *iph = skb->nh.iph;
 	__u16 *portp = (__u16 *)&(((char *)iph)[iph->ihl*4]);
 
-	/*
+        /* if it is fwmark-based service, the cache_bypass sysctl is up
+           and the destination is RTN_UNICAST (and not local), then create
+           a cache_bypass connection entry */
+        if (sysctl_ip_vs_cache_bypass && svc->fwmark
+            && (inet_addr_type(iph->daddr) == RTN_UNICAST)) {
+                int ret;
+                struct ip_vs_conn *cp;
+                
+                ip_vs_service_put(svc);
+
+                /* create a new connection entry */
+                IP_VS_DBG(6, "ip_vs_leave: create a cache_bypass entry\n");
+                cp = ip_vs_conn_new(iph->protocol,
+                                    iph->saddr, portp[0],
+                                    iph->daddr, portp[1],
+                                    0, 0, 0);
+                if (cp == NULL) {
+                        IP_VS_ERR("ip_vs_conn_new failed\n");
+                        kfree_skb(skb);
+                        return NF_STOLEN;
+                }
+
+                /* bind the bypass_xmit */
+                ip_vs_bind_bypass_xmit(cp);
+
+                /* set state */
+                ip_vs_set_state(cp, VS_STATE_INPUT, iph, portp); 
+
+                /* transmit the first SYN packet */
+                ret = cp->packet_xmit(skb, cp);
+                
+                ip_vs_conn_put(cp);
+                return ret;
+        }
+
+        /*
 	 * When the virtual ftp service is presented, packets destined
 	 * for other services on the VIP may get here (except services
 	 * listed in the ipvs table), pass the packets, because it is
 	 * not ipvs job to decide to drop the packets.
 	 */
 	if ((svc->port == FTPPORT) && (portp[1] != FTPPORT)) {
-		atomic_dec(&svc->refcnt);
+                ip_vs_service_put(svc);
 		return NF_ACCEPT;
 	}
-	atomic_dec(&svc->refcnt);
+
+        ip_vs_service_put(svc);
 
 	/*
 	 * Notify the client that the destination is unreachable, and
@@ -463,430 +499,6 @@ static int ip_vs_leave(struct ip_vs_service *svc, struct sk_buff *skb)
 }
 
 
-/*
- *      NAT transmitter (only for outside-to-inside nat forwarding)
- */
-static int ip_vs_nat_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
-{
-        struct rtable *rt;			/* Route to the other host */
-	struct ip_vs_dest *dest;
-	struct iphdr  *iph = skb->nh.iph;
-	u8     tos = iph->tos;
-	u32    rtos;
-	union ip_vs_tphdr h;
-	unsigned short size;
-	int doff;
-	int csum = 0;
-	int csum_ok = 0;
-        int mtu;
-
-	EnterFunction(4);
-
-        h.raw = (char*) iph + iph->ihl * 4;
-	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-	doff = proto_doff(iph->protocol, h.raw, size);
-
-        /*
-         *  Must do TCP/UDP checksum checking before mangling the packet
-         */
-        if ((iph->protocol != IPPROTO_UDP) || (h.uh->check != 0)) {
-                switch (skb->ip_summed) {
-                case CHECKSUM_NONE:
-                        csum = csum_partial(h.raw + doff, size - doff, 0);
-                        csum_ok++;
-                        skb->csum = csum_partial(h.raw , doff, csum);
-
-                case CHECKSUM_HW:
-                        if (csum_tcpudp_magic(iph->saddr, iph->daddr,
-                                              size, iph->protocol, skb->csum))
-                        {
-                                /*  atomic_dec(&svc->refcnt); */
-                                IP_VS_DBG(0, "Incoming failed %s checksum "
-                                          "from %d.%d.%d.%d (size=%d)!\n",
-                                          vs_proto_name(iph->protocol),
-                                          NIPQUAD(iph->saddr),
-                                          size);
-                                kfree_skb(skb);
-                                return -EINVAL;
-                        }
-                default:/* CHECKSUM_UNNECESSARY */
-                }
-        }
-        
-        /*
-         *  Check if it is no_cport connection (VS/NAT only)...
-         */
-        if (cp->flags & IP_VS_CONN_F_NO_CPORT ) {
-                atomic_dec(&ip_vs_conn_no_cport_cnt);
-                ip_vs_conn_unhash(cp);
-                cp->flags &= ~IP_VS_CONN_F_NO_CPORT;
-                cp->cport = h.portp[0];
-                /* hash on new dport */
-                ip_vs_conn_hash(cp);
-
-                IP_VS_DBG(1, "filled cport=%d\n", ntohs(cp->dport));
-        }
-                
-	dest = cp->dest;
-	rtos = RT_TOS(tos);
-	spin_lock(&dest->dst_lock);
-	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
-		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
-			spin_unlock(&dest->dst_lock);
-			IP_VS_DBG(0, "ip_vs_nat_xmit(): route error, "
-				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
-			goto tx_error_icmp;
-		}
-		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
-		IP_VS_DBG(1, "NAT: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
-			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
-			rtos);
-	}
-	spin_unlock(&dest->dst_lock);
-
-        /* MTU checking */
-        mtu = rt->u.dst.pmtu;
-        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
-                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
-                ip_rt_put(rt);
-                IP_VS_DBG(0, "ip_vs_nat_xmit(): frag needed\n");
-                goto tx_error;
-        }
-        
-	/* drop old route */
-        dst_release(skb->dst);
-	skb->dst = &rt->u.dst;
-
-        /* mangle the packet */
-        iph->daddr = cp->daddr;
-        h.portp[1] = cp->dport;
-
-        /*
-         *	Invalidate csum saving if it has ip_vs_app helper
-         */
-        if (cp->app)
-                csum_ok = 0;
-
-        /*
-         *	Attempt ip_vs_app call.
-         *	will fix ip_vs_conn and iph ack_seq stuff
-         */
-        if (ip_vs_app_pkt_in(cp, &skb) != 0) {
-                /*
-                 *	skb has changed, update pointers.
-                 */
-                /*  skb = *skb_p; */
-                iph = skb->nh.iph;
-                h.raw = (char*) iph + iph->ihl*4;
-                size = ntohs(iph->tot_len) - (iph->ihl * 4);
-        }
-
-        /*
-         * Yug! adjust UDP/TCP checksums
-         */
-        /*
-         *	Transport's payload partial csum
-         */
-        if (!csum_ok) {
-                csum = csum_partial(h.raw + doff, size - doff, 0);
-        }
-        skb->csum = csum;
-
-        /*
-         *	Protocol csum
-         */
-        switch (iph->protocol) {
-        case IPPROTO_TCP:
-                h.th->check = 0;
-                h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-                                              size, iph->protocol,
-                                              csum_partial(h.raw, doff, csum));
-                break;
-        case IPPROTO_UDP:
-                h.uh->check = 0;
-                h.uh->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
-                                              size, iph->protocol,
-                                              csum_partial(h.raw, doff, csum));
-                if (h.uh->check == 0)
-                        h.uh->check = 0xFFFF;
-                break;
-        }
-        ip_send_check(iph);
-
-        IP_VS_DBG(1, "NAT to %u.%u.%u.%u:%d\n",
-                  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
-
-        /* FIXME: when application helper enlarges the packet and the length
-           is larger than the MTU of outgoing device, there will be still
-           MTU problem. */
-
-#ifdef CONFIG_NETFILTER_DEBUG
-        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
-#endif /* CONFIG_NETFILTER_DEBUG */
-        ip_send(skb);
-
-        LeaveFunction(4);
-	return 0;
-
-  tx_error_icmp:
-	dst_link_failure(skb);
-  tx_error:
-	kfree_skb(skb);
-	return -EINVAL;
-}
-
-
-/*
- *   This function encapsulates the packet in a new IP header, its destination
- *   will be set to the daddr. Most code of this function is from ipip.c.
- *   Usage:
- *     It is called in the ip_vs_forward() function. The load balancer
- *     selects a real server from a cluster based on a scheduling algorithm,
- *     encapsulates the packet and forwards it to the selected server. All real
- *     servers are configured with "ifconfig tunl0 <Virtual IP Address> up".
- *     When the server receives the encapsulated packet, it decapsulates the
- *     packet, processes the request and return the reply packets directly to
- *     the client without passing the load balancer. This can greatly
- *     increase the scalability of virtual server.
- *   Returns:
- *     if succeeded, return 0; otherwise, return -EINVAL.
- */
-static int ip_vs_tunnel_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
-{
-	struct rtable *rt;			/* Route to the other host */
-	struct ip_vs_dest *dest;
-	struct net_device *tdev;		/* Device to other host */
-	struct iphdr  *old_iph = skb->nh.iph;
-	u8     tos = old_iph->tos;
-	u32    rtos;
-	u16    df = old_iph->frag_off;
-	struct iphdr  *iph;			/* Our new IP header */
-	int    max_headroom;			/* The extra header space needed */
-	int    mtu;
-
-	EnterFunction(4);
-
-        if (skb->protocol != __constant_htons(ETH_P_IP)) {
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): protocol error, "
-                          "ETH_P_IP: %d, skb protocol: %d\n",
-			  __constant_htons(ETH_P_IP), skb->protocol);
-		goto tx_error;
-	}
-
-	dest = cp->dest;
-	rtos = RT_TOS(tos);
-	spin_lock(&dest->dst_lock);
-	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
-		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
-			spin_unlock(&dest->dst_lock);
-			IP_VS_DBG(0, "ip_vs_tunnel_xmit(): route error, "
-				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
-			goto tx_error_icmp;
-		}
-		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
-		IP_VS_DBG(1, "TUN: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
-			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
-			rtos);
-	}
-	spin_unlock(&dest->dst_lock);
-
-	tdev = rt->u.dst.dev;
-
-	mtu = rt->u.dst.pmtu - sizeof(struct iphdr);
-	if (mtu < 68) {
-		ip_rt_put(rt);
-		IP_VS_DBG(0, "ip_vs_tunnel_xmit(): mtu less than 68\n");
-		goto tx_error;
-	}
-	if (skb->dst && mtu < skb->dst->pmtu)
-		skb->dst->pmtu = mtu;
-
-	df |= (old_iph->frag_off&__constant_htons(IP_DF));
-
-	if ((old_iph->frag_off&__constant_htons(IP_DF))
-            && mtu < ntohs(old_iph->tot_len)) {
-		icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
-		ip_rt_put(rt);
-		IP_VS_DBG(1, "ip_vs_tunnel_xmit(): frag needed\n");
-		goto tx_error;
-	}
-
-	skb->h.raw = skb->nh.raw;
-
-	/*
-	 * Okay, now see if we can stuff it in the buffer as-is.
-	 */
-	max_headroom = (((tdev->hard_header_len+15)&~15)+sizeof(struct iphdr));
-
-	if (skb_headroom(skb) < max_headroom
-            || skb_cloned(skb) || skb_shared(skb)) {
-		struct sk_buff *new_skb =
-                        skb_realloc_headroom(skb, max_headroom);
-		if (!new_skb) {
-			ip_rt_put(rt);
-			kfree_skb(skb);
-			IP_VS_ERR("ip_vs_tunnel_xmit(): no memory\n");
-			return -EINVAL;
-		}
-		kfree_skb(skb);
-		skb = new_skb;
-	}
-
-	skb->nh.raw = skb_push(skb, sizeof(struct iphdr));
-	memset(&(IPCB(skb)->opt), 0, sizeof(IPCB(skb)->opt));
-
-	/* drop old route */
-	dst_release(skb->dst);
-	skb->dst = &rt->u.dst;
-
-	/*
-	 *	Push down and install the IPIP header.
-	 */
-	iph			=	skb->nh.iph;
-	iph->version		=	4;
-	iph->ihl		=	sizeof(struct iphdr)>>2;
-	iph->frag_off		=	df;
-	iph->protocol		=	IPPROTO_IPIP;
-	iph->tos		=	tos;
-	iph->daddr		=	rt->rt_dst;
-	iph->saddr		=	rt->rt_src;
-	iph->ttl		=	old_iph->ttl;
-	iph->tot_len		=	htons(skb->len);
-	ip_select_ident(iph, &rt->u.dst);
-	ip_send_check(iph);
-
-#ifdef CONFIG_NETFILTER_DEBUG
-        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
-#endif /* CONFIG_NETFILTER_DEBUG */
-	ip_send(skb);
-
-        LeaveFunction(4);
-        
-	return 0;
-
-  tx_error_icmp:
-	dst_link_failure(skb);
-  tx_error:
-	dev_kfree_skb(skb);
-	return -EINVAL;
-}
-
-
-#if 0000
-/* Need this wrapper because NF_HOOK takes the function address */
-static inline int do_ip_send(struct sk_buff *skb)
-{
-        return ip_send(skb);
-}
-#endif
-
-/*
- *      Direct Routing
- */
-static int ip_vs_dr_xmit(struct sk_buff *skb, struct ip_vs_conn *cp)
-{
-	struct rtable *rt;			/* Route to the other host */
-	struct ip_vs_dest *dest;
-	struct iphdr  *iph = skb->nh.iph;
-	u8     tos = iph->tos;
-	u32    rtos;
-        int    mtu;
-
-	EnterFunction(4);
-
-	dest = cp->dest;
-	rtos = RT_TOS(tos);
-	spin_lock(&dest->dst_lock);
-	if (!sysctl_use_dst || !(rt = (struct rtable *) __ip_vs_dst_check(dest, rtos, 0))) {
-		if (ip_route_output(&rt, dest->addr, 0, rtos, 0)) {
-			spin_unlock(&dest->dst_lock);
-			IP_VS_DBG(0, "ip_vs_dr_xmit(): route error, "
-				  "dest: %u.%u.%u.%u\n", NIPQUAD(dest->addr));
-			goto tx_error_icmp;
-		}
-		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
-		IP_VS_DBG(1, "DR: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
-			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
-			rtos);
-	}
-	spin_unlock(&dest->dst_lock);
-
-        /* MTU checking */
-        mtu = rt->u.dst.pmtu;
-        if ((skb->len > mtu) && (iph->frag_off&__constant_htons(IP_DF))) {
-                icmp_send(skb, ICMP_DEST_UNREACH,ICMP_FRAG_NEEDED, htonl(mtu));
-                ip_rt_put(rt);
-                IP_VS_DBG(0, "ip_vs_dr_xmit(): frag needed\n");
-                goto tx_error;
-        }
-
-	/* drop old route */
-        dst_release(skb->dst);
-	skb->dst = &rt->u.dst;
-
-#ifdef CONFIG_NETFILTER_DEBUG
-        skb->nf_debug = 1 << NF_IP_LOCAL_OUT;
-#endif /* CONFIG_NETFILTER_DEBUG */
-        ip_send(skb);
-
-#if 0000
-        NF_HOOK(PF_INET, NF_IP_LOCAL_OUT, skb, NULL, rt->u.dst.dev,
-                do_ip_send);
-#endif
-        LeaveFunction(4);
-	return 0;
-
-  tx_error_icmp:
-	dst_link_failure(skb);
-  tx_error:
-	kfree_skb(skb);
-	return -EINVAL;
-}
-
-
-/*
- *      ip_vs_forward forwards the packet through tunneling, direct
- *      routing or local node (passing to the upper layer).
- *      in the outside-to-inside direction.
- */
-static inline int ip_vs_forward(struct sk_buff *skb, struct ip_vs_conn *cp)
-{
-        atomic_inc(&cp->in_pkts);
-        
-        switch (IP_VS_FWD_METHOD(cp)) {
-        case IP_VS_CONN_F_MASQ:
-                if (ip_vs_nat_xmit(skb, cp)) {
-                        IP_VS_DBG(2, "nat failed.\n");
-                } else {
-                        IP_VS_DBG(2, "nat succeeded.\n");
-                }
-                return NF_STOLEN;
-                
-        case IP_VS_CONN_F_TUNNEL:
-                if (ip_vs_tunnel_xmit(skb, cp)) {
-                        IP_VS_DBG(2, "tunneling failed.\n");
-                } else {
-                        IP_VS_DBG(2, "tunneling succeeded.\n");
-                }
-                return NF_STOLEN;
-                                
-        case IP_VS_CONN_F_DROUTE:
-                if (ip_vs_dr_xmit(skb, cp)) {
-                        IP_VS_DBG(2, "direct routing failed.\n");
-                } else {
-                        IP_VS_DBG(2, "direct routing succeeded.\n");
-                }
-                return NF_STOLEN;
-                                
-        case IP_VS_CONN_F_LOCALNODE:
-                IP_VS_DBG(2, "local node forwarding.\n");
-                return NF_ACCEPT;
-        }
-        
-        return NF_ACCEPT;
-}
-
-
 /*
  *	Handle ICMP messages in the inside-to-outside direction (outgoing).
  *	Find any that might be relevant, check against existing connections,
@@ -904,7 +516,7 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	struct ip_vs_conn *cp;
 	unsigned short	 len   = ntohs(iph->tot_len) - (iph->ihl * 4);
 
-	IP_VS_DBG(3, "outgoing ICMP (%d,%d) %u.%u.%u.%u->%u.%u.%u.%u\n",
+	IP_VS_DBG(12, "outgoing ICMP (%d,%d) %u.%u.%u.%u->%u.%u.%u.%u\n",
 		  icmph->type, ntohs(icmp_id(icmph)),
 		  NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
 
@@ -944,7 +556,7 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 		return NF_DROP;
 	}
 
-	IP_VS_DBG(1, "Handling outgoing ICMP for "
+	IP_VS_DBG(11, "Handling outgoing ICMP for "
                   "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
 		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
 		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
@@ -977,7 +589,7 @@ static int ip_vs_out_icmp(struct sk_buff **skb_p)
 	icmph->checksum = 0;
 	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
 
-	IP_VS_DBG(1, "Forwarding correct outgoing ICMP to "
+	IP_VS_DBG(11, "Forwarding correct outgoing ICMP to "
                   "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
 		  NIPQUAD(ciph->saddr), ntohs(pptr[0]),
 		  NIPQUAD(ciph->daddr), ntohs(pptr[1]));
@@ -1012,11 +624,11 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	int csum = 0;
 	int csum_ok = 0;
 
-	EnterFunction(4);
+	EnterFunction(11);
 
 	h.raw = (char*) iph + iph->ihl * 4;
 	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-	doff = proto_doff(iph->protocol, h.raw, size);
+	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
 
 	switch (doff) {
 	case 0:
@@ -1033,8 +645,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	if (iph->protocol == IPPROTO_ICMP)
 		return ip_vs_out_icmp(skb_p);
 
-        IP_VS_DBG(3, "O-pkt: %s size=%d\n",
-                  vs_proto_name(iph->protocol), size);
+        IP_VS_DBG(11, "O-pkt: %s size=%d\n",
+                  ip_vs_proto_name(iph->protocol), size);
                 
         /*
          *	Now hunt the list to see if we have an old entry
@@ -1055,9 +667,9 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 				return NF_STOLEN;
 			}
 		}
-                IP_VS_DBG(2, "packet for %s %d.%d.%d.%d:%d "
+                IP_VS_DBG(12, "packet for %s %d.%d.%d.%d:%d "
                           "continue traversal as normal.\n",
-                          vs_proto_name(iph->protocol),
+                          ip_vs_proto_name(iph->protocol),
                           NIPQUAD(iph->daddr),
                           ntohs(h.portp[1]));
                 return NF_ACCEPT;
@@ -1069,8 +681,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                 {
                 case CHECKSUM_NONE:
                         csum = csum_partial(h.raw + doff, size - doff, 0);
-                        IP_VS_DBG(3, "O-pkt: %s I-datacsum=%d\n",
-                                  vs_proto_name(iph->protocol), csum);
+                        IP_VS_DBG(11, "O-pkt: %s I-datacsum=%d\n",
+                                  ip_vs_proto_name(iph->protocol), csum);
                         skb->csum = csum_partial(h.raw , doff, csum);
                         break;
 
@@ -1079,7 +691,7 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                                               iph->protocol, skb->csum)) {
                                 IP_VS_DBG(0, "Outgoing failed %s checksum "
                                           "from %d.%d.%d.%d (size=%d)!\n",
-                                          vs_proto_name(iph->protocol),
+                                          ip_vs_proto_name(iph->protocol),
                                           NIPQUAD(iph->saddr),
                                           size);
                                 return NF_DROP;
@@ -1089,8 +701,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                 }
         }
 
-        IP_VS_DBG(2, "Outgoing %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
-                  vs_proto_name(iph->protocol),
+        IP_VS_DBG(11, "Outgoing %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
+                  ip_vs_proto_name(iph->protocol),
                   NIPQUAD(iph->saddr), ntohs(h.portp[0]),
                   NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
@@ -1137,8 +749,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
 	}
 	skb->csum = csum;
 
-	IP_VS_DBG(3, "O-pkt: %s size=%d O-datacsum=%d\n",
-                  vs_proto_name(iph->protocol), size, csum);
+	IP_VS_DBG(11, "O-pkt: %s size=%d O-datacsum=%d\n",
+                  ip_vs_proto_name(iph->protocol), size, csum);
 
 	/*
 	 * 	Protocol csum
@@ -1149,8 +761,8 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                 h.th->check=csum_tcpudp_magic(iph->saddr, iph->daddr,
                                               size, iph->protocol,
                                               csum_partial(h.raw, doff, csum));
-                IP_VS_DBG(3, "O-pkt: %s O-csum=%d (+%d)\n",
-                          vs_proto_name(iph->protocol), h.th->check,
+                IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
+                          ip_vs_proto_name(iph->protocol), h.th->check,
                           (char*) & (h.th->check) - (char*) h.raw);
                 break;
         case IPPROTO_UDP:
@@ -1160,17 +772,17 @@ static unsigned int ip_vs_out(unsigned int hooknum,
                                               csum_partial(h.raw, doff, csum));
                 if (h.uh->check == 0)
                         h.uh->check = 0xFFFF;
-                IP_VS_DBG(3, "O-pkt: %s O-csum=%d (+%d)\n",
-                          vs_proto_name(iph->protocol), h.uh->check,
+                IP_VS_DBG(11, "O-pkt: %s O-csum=%d (+%d)\n",
+                          ip_vs_proto_name(iph->protocol), h.uh->check,
                           (char*) &(h.uh->check)- (char*) h.raw);
                 break;
 	}
 	ip_send_check(iph);
 
-	vs_set_state(cp, VS_STATE_OUTPUT, iph, h.portp);
+	ip_vs_set_state(cp, VS_STATE_OUTPUT, iph, h.portp);
 	ip_vs_conn_put(cp);
         
-        LeaveFunction(4);
+        LeaveFunction(11);
 	return NF_ACCEPT;
 }
 
@@ -1196,7 +808,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	u32    rtos;
         int    mtu;
 
- 	IP_VS_DBG(2, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
+ 	IP_VS_DBG(12, "icmp in (%d,%d) %u.%u.%u.%u -> %u.%u.%u.%u\n",
                   icmph->type, ntohs(icmp_id(icmph)),
                   NIPQUAD(iph->saddr), NIPQUAD(iph->daddr));
 
@@ -1227,13 +839,13 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 
         pptr = (__u16 *)&(((char *)ciph)[ciph->ihl*4]);
 
- 	IP_VS_DBG(2, "Handling incoming ICMP for "
+ 	IP_VS_DBG(11, "Handling incoming ICMP for "
                   "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
                   NIPQUAD(ciph->saddr), ntohs(pptr[0]),
                   NIPQUAD(ciph->daddr), ntohs(pptr[1]));
 
-	/* This is pretty much what ip_vs_conn_in_get_iph() does,
-           except paracp are wrong way round */
+	/* This is pretty much what ip_vs_conn_in_get() does,
+           except parameters are wrong way round */
         cp = ip_vs_conn_in_get(ciph->protocol,
                                ciph->daddr, pptr[1],
                                ciph->saddr, pptr[0]);
@@ -1241,9 +853,14 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 		return NF_ACCEPT;
 
         /* The ICMP packet for VS/TUN, VS/DR and LOCALNODE will be
-           forwarded directly here */
+           forwarded directly here, because there is no need to
+           translate address/port back */
         if (IP_VS_FWD_METHOD(cp) != IP_VS_CONN_F_MASQ) {
-                int ret = ip_vs_forward(skb, cp);
+                int ret;
+                if (cp->packet_xmit)
+                        ret = cp->packet_xmit(skb, cp);
+                else
+                        ret = NF_ACCEPT;
                 ip_vs_conn_put(cp);
                 return ret;
         }
@@ -1275,7 +892,7 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 	icmph->checksum = 0;
 	icmph->checksum = ip_compute_csum((unsigned char *) icmph, len);
 
- 	IP_VS_DBG(2, "Forwarding incoming ICMP to "
+ 	IP_VS_DBG(11, "Forwarding incoming ICMP to "
                   "%u.%u.%u.%u:%d -> %u.%u.%u.%u:%d\n",
                   NIPQUAD(ciph->saddr), ntohs(pptr[0]),
                   NIPQUAD(ciph->daddr), ntohs(pptr[1]));
@@ -1294,9 +911,9 @@ static int ip_vs_in_icmp(struct sk_buff **skb_p)
 			goto tx_error_icmp;
 		}
 		__ip_vs_dst_set(dest, rtos, dst_clone(&rt->u.dst));
-		IP_VS_DBG(1, "ICMP: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
-			NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
-			rtos);
+		IP_VS_DBG(10, "ICMP: new dst %u.%u.%u.%u, refcnt=%d, rtos=%X\n",
+                          NIPQUAD(dest->addr), atomic_read(&rt->u.dst.__refcnt),
+                          rtos);
         }
 	spin_unlock(&dest->dst_lock);
 
@@ -1350,7 +967,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	 *	... don't know why 1st test DOES NOT include 2nd (?)
 	 */
 	if (skb->pkt_type != PACKET_HOST || skb->dev == &loopback_dev) {
-		IP_VS_DBG(4, "packet type=%d proto=%d daddr=%d.%d.%d.%d ignored\n",
+		IP_VS_DBG(12, "packet type=%d proto=%d daddr=%d.%d.%d.%d ignored\n",
 			  skb->pkt_type,
 			  iph->protocol,
 			  NIPQUAD(iph->daddr));
@@ -1363,7 +980,7 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	 *	IP payload size
 	 */
 	size = ntohs(iph->tot_len) - (iph->ihl * 4);
-	doff = proto_doff(iph->protocol, h.raw, size);
+	doff = ip_vs_proto_doff(iph->protocol, h.raw, size);
 
 	switch (doff) {
 	case 0:
@@ -1380,8 +997,8 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 	if (iph->protocol == IPPROTO_ICMP)
 		return(ip_vs_in_icmp(skb_p));
 
-	IP_VS_DBG(2, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
-		  vs_proto_name(iph->protocol),
+	IP_VS_DBG(11, "Incoming %s %u.%u.%u.%u:%d->%u.%u.%u.%u:%d\n",
+		  ip_vs_proto_name(iph->protocol),
 		  NIPQUAD(iph->saddr), ntohs(h.portp[0]),
 		  NIPQUAD(iph->daddr), ntohs(h.portp[1]));
 
@@ -1406,15 +1023,15 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 
 	if (!cp
             && (h.th->syn || (iph->protocol!=IPPROTO_TCP))
-            && (svc = ip_vs_lookup_service(skb->nfmark, iph->protocol,
-                                           iph->daddr, h.portp[1]))) {
+            && (svc = ip_vs_service_get(skb->nfmark, iph->protocol,
+                                        iph->daddr, h.portp[1]))) {
                 
 		if (ip_vs_todrop()) {
 			/*
 			 * It seems that we are very loaded.
 			 * We have to drop this packet :(
 			 */
-			atomic_dec(&svc->refcnt);
+                        ip_vs_service_put(svc);
 			return NF_DROP;
 		}
                 
@@ -1425,22 +1042,27 @@ static unsigned int ip_vs_in(unsigned int hooknum,
 		cp = ip_vs_schedule(svc, iph);
 		if (!cp)
 			return ip_vs_leave(svc, skb);
-                atomic_dec(&svc->refcnt);
+                ip_vs_service_put(svc);
 	}
 
 	if (cp != NULL) {
                 int ret;
 
-                vs_set_state(cp, VS_STATE_INPUT, iph, h.portp); 
-                ret = ip_vs_forward(skb, cp);
+                ip_vs_set_state(cp, VS_STATE_INPUT, iph, h.portp); 
+                if (cp->packet_xmit)
+                        ret = cp->packet_xmit(skb, cp);
+                else {
+                        IP_VS_DBG(0, "warning: packet_xmit is null");
+                        ret = NF_ACCEPT;
+                }
                 ip_vs_conn_put(cp);
                 return ret;
 	}
 
 	/* sorry, all this trouble for a no-hit :) */
-        IP_VS_DBG(3, "packet for %s %d.%d.%d.%d:%d continue traversal "
+        IP_VS_DBG(12, "packet for %s %d.%d.%d.%d:%d continue traversal "
                   "as normal.\n",
-                  vs_proto_name(iph->protocol),
+                  ip_vs_proto_name(iph->protocol),
                   NIPQUAD(iph->daddr),
                   ntohs(h.portp[1]));
 
@@ -1470,8 +1092,6 @@ static int __init ip_vs_init(void)
 {
 	int ret;
 
-	EnterFunction(4);
-
         ret = ip_vs_control_init();
 	if (ret < 0) {
 		IP_VS_ERR("can't setup control.\n");
@@ -1504,9 +1124,6 @@ static int __init ip_vs_init(void)
 	}
 
         IP_VS_INFO("ipvs loaded.\n");
-
-        LeaveFunction(4);
-        
 	return ret;
 
   cleanup_inops:
