@@ -5,7 +5,7 @@
  *              high-performance and highly available server based on a
  *              cluster of servers.
  *
- * Version:     $Id: ip_vs_ctl.c,v 1.7 2000/12/19 15:04:23 wensong Exp $
+ * Version:     $Id: ip_vs_ctl.c,v 1.9 2001/01/10 07:42:53 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *              Peter Kese <peter.kese@ijs.si>
@@ -73,6 +73,7 @@ static int sysctl_ip_vs_secure_tcp = 0;
 static int sysctl_ip_vs_amemthresh = 1024;
 static int sysctl_ip_vs_am_droprate = 10;
 static int sysctl_use_dst = 1;
+static int sysctl_ip_vs_cache_bypass = 0;
 
 #ifdef CONFIG_IP_VS_DEBUG
 static int sysctl_ip_vs_debug_level = 0;
@@ -301,10 +302,10 @@ static int ip_vs_svc_unhash(struct ip_vs_service *svc)
 
 
 /*
- *  Lookup service by {proto,addr,port} in the service table.
+ *  Get service by {proto,addr,port} in the service table.
  */
 static __inline__ struct ip_vs_service *
-__ip_vs_lookup_service(__u16 protocol, __u32 vaddr, __u16 vport)
+__ip_vs_service_get(__u16 protocol, __u32 vaddr, __u16 vport)
 {
 	unsigned hash;
 	struct ip_vs_service *svc;
@@ -332,10 +333,10 @@ __ip_vs_lookup_service(__u16 protocol, __u32 vaddr, __u16 vport)
 
 
 /*
- *  Lookup service by fwmark in the service table.
+ *  Get service by {fwmark} in the service table.
  */
 static __inline__ struct ip_vs_service *
-__ip_vs_lookup_svc_fwm(unsigned long fwmark)
+__ip_vs_svc_fwm_get(unsigned long fwmark)
 {
 	unsigned hash;
 	struct ip_vs_service *svc;
@@ -360,7 +361,8 @@ __ip_vs_lookup_svc_fwm(unsigned long fwmark)
 }
 
 struct ip_vs_service *
-ip_vs_lookup_service(unsigned long fwmark, __u16 protocol, __u32 vaddr, __u16 vport)
+ip_vs_service_get(unsigned long fwmark,
+                  __u16 protocol, __u32 vaddr, __u16 vport)
 {
 	struct ip_vs_service *svc;
 
@@ -370,13 +372,13 @@ ip_vs_lookup_service(unsigned long fwmark, __u16 protocol, __u32 vaddr, __u16 vp
 		/*	
 		 *	Check the table hashed by fwmark
 		 */
-		svc = __ip_vs_lookup_svc_fwm(fwmark);
+		svc = __ip_vs_svc_fwm_get(fwmark);
 	} else {
 		/*	
 		 *	Check the table hashed by <protocol,addr,port>
 		 *	first for "full" addressed entries
 		 */
-		svc = __ip_vs_lookup_service(protocol, vaddr, vport);
+		svc = __ip_vs_service_get(protocol, vaddr, vport);
 
 		if (svc == NULL
                     && protocol == IPPROTO_TCP
@@ -386,7 +388,7 @@ ip_vs_lookup_service(unsigned long fwmark, __u16 protocol, __u32 vaddr, __u16 vp
 			 * Check if ftp service entry exists, the packet
 			 * might belong to FTP data connections.
 			 */
-			svc = __ip_vs_lookup_service(protocol, vaddr, FTPPORT);
+			svc = __ip_vs_service_get(protocol, vaddr, FTPPORT);
 		}
 
 		if (svc == NULL
@@ -394,14 +396,14 @@ ip_vs_lookup_service(unsigned long fwmark, __u16 protocol, __u32 vaddr, __u16 vp
 			/*
 			 * Check if the catch-all port (port zero) exists
 			 */
-			svc = __ip_vs_lookup_service(protocol, vaddr, 0);
+			svc = __ip_vs_service_get(protocol, vaddr, 0);
 		}
 	}
 
 	read_unlock(&__ip_vs_svc_lock);
 
-        IP_VS_DBG(2, "lookup_service fwm %ld %s %u.%u.%u.%u:%u %s\n",
-                  fwmark, vs_proto_name(protocol),
+        IP_VS_DBG(6, "lookup service: fwm %ld %s %u.%u.%u.%u:%u %s\n",
+                  fwmark, ip_vs_proto_name(protocol),
                   NIPQUAD(vaddr), ntohs(vport),
                   svc?"hit":"not hit");
 
@@ -545,7 +547,7 @@ ip_vs_get_trash_dest(struct ip_vs_service *svc, __u32 daddr, __u16 dport)
 
 	for (e=l->next; e!=l; e=e->next) {
 		dest = list_entry(e, struct ip_vs_dest, n_list);
-		IP_VS_DBG(1, "Destination %lu/%u.%u.%u.%u:%u still in trash, "
+		IP_VS_DBG(3, "Destination %lu/%u.%u.%u.%u:%u still in trash, "
                           "refcnt=%d\n",
                           dest->vfwmark,
                           NIPQUAD(dest->addr), ntohs(dest->port),
@@ -565,7 +567,7 @@ ip_vs_get_trash_dest(struct ip_vs_service *svc, __u32 daddr, __u16 dport)
 		 * Try to purge the destination from trash if not referenced
 		 */
 		if (atomic_read(&dest->refcnt) == 1) {
-			IP_VS_DBG(1, "Removing destination %lu/%u.%u.%u.%u:%u "
+			IP_VS_DBG(3, "Removing destination %lu/%u.%u.%u.%u:%u "
                                   "from trash\n",
 				  dest->vfwmark,
 				  NIPQUAD(dest->addr), ntohs(dest->port));
@@ -615,9 +617,9 @@ static int __ip_vs_update_dest(struct ip_vs_service *svc,
 	 */
 	if ((conn_flags & IP_VS_CONN_F_FWD_MASK) != 0) {
 		conn_flags |= IP_VS_CONN_F_NOOUTPUT;
-		atomic_set(&dest->conn_flags,conn_flags);
+		atomic_set(&dest->conn_flags, conn_flags);
 	} else {
-		atomic_set(&dest->conn_flags,conn_flags);
+		atomic_set(&dest->conn_flags, conn_flags);
 		/*
 		 *    Put the real service in ip_vs_rtable if not present.
 		 *    For now only for NAT!
@@ -646,7 +648,7 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
 	struct ip_vs_dest *dest;
 	int ret;
 
-	EnterFunction(4);
+	EnterFunction(2);
 
 	*destp = dest = (struct ip_vs_dest*) kmalloc(sizeof(struct ip_vs_dest),
 					GFP_ATOMIC);
@@ -672,7 +674,7 @@ ip_vs_new_dest(struct ip_vs_service *svc, struct ip_vs_rule_user *ur,
 	ret = __ip_vs_update_dest(svc, dest, ur);
 	if (ret) kfree(dest);
 
-	LeaveFunction(4);
+	LeaveFunction(2);
 
 	return ret;
 }
@@ -689,7 +691,7 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 	__u16 dport = ur->dport;
 	int ret;
 
-	EnterFunction(4);
+	EnterFunction(2);
 
 	if (ur->weight < 0) {
 		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
@@ -711,7 +713,7 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 	 */
 	dest = ip_vs_get_trash_dest(svc, daddr, dport);
 	if (dest != NULL) {
-		IP_VS_DBG(1, "Get destination %u.%u.%u.%u:%u from trash, "
+		IP_VS_DBG(3, "Get destination %u.%u.%u.%u:%u from trash, "
 			"refcnt=%d, service %lu/%u.%u.%u.%u:%u\n",
 			NIPQUAD(daddr), ntohs(dport),
 			atomic_read(&dest->refcnt),
@@ -765,7 +767,7 @@ static int ip_vs_add_dest(struct ip_vs_service *svc,
 
 	write_unlock_bh(&__ip_vs_svc_lock);
 
-        LeaveFunction(4);
+        LeaveFunction(2);
         
 	return 0;
 }
@@ -782,7 +784,7 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 	__u16 dport = ur->dport;
 	int ret;
 
-	EnterFunction(4);
+	EnterFunction(2);
 
 	if (ur->weight < 0) {
 		IP_VS_ERR("ip_vs_add_dest(): server weight less than zero\n");
@@ -800,7 +802,7 @@ static int ip_vs_edit_dest(struct ip_vs_service *svc,
 
 	ret = __ip_vs_update_dest(svc, dest, ur);
 
-        LeaveFunction(4);
+        LeaveFunction(2);
         
 	return ret;
 }
@@ -828,7 +830,7 @@ static void __ip_vs_del_dest(struct ip_vs_dest *dest)
 		__ip_vs_dst_reset(dest);
 		kfree(dest);
 	} else {
-		IP_VS_DBG(1, "Moving dest %u.%u.%u.%u:%u into trash, refcnt=%d\n",
+		IP_VS_DBG(3, "Moving dest %u.%u.%u.%u:%u into trash, refcnt=%d\n",
                           NIPQUAD(dest->addr), ntohs(dest->port),
                           atomic_read(&dest->refcnt));
 		list_add(&dest->n_list, &ip_vs_dest_trash);
@@ -865,7 +867,7 @@ static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 	__u32 daddr = ur->daddr;
 	__u16 dport = ur->dport;
 
-	EnterFunction(4);
+	EnterFunction(2);
 
 	/*
 	 *    Lookup the destination list
@@ -894,7 +896,7 @@ static int ip_vs_del_dest(struct ip_vs_service *svc,struct ip_vs_rule_user *ur)
 	 */
 	__ip_vs_del_dest(dest);
 
-        LeaveFunction(4);
+        LeaveFunction(2);
 
 	return 0;
 }
@@ -1205,7 +1207,7 @@ static int ip_vs_sysctl_defense_mode(ctl_table *ctl, int write,
  */
 struct ip_vs_sysctl_table {
 	struct ctl_table_header *sysctl_header;
-	ctl_table vs_vars[20];
+	ctl_table vs_vars[21];
 	ctl_table vs_dir[2];
 	ctl_table ipv4_dir[2];
 	ctl_table root_dir[2];
@@ -1273,6 +1275,9 @@ static struct ip_vs_sysctl_table ipv4_vs_table = {
          {NET_IPV4_VS_TO_ICMP, "timeout_icmp",
           &vs_timeout_table_dos.timeout[IP_VS_S_ICMP],
           sizeof(int), 0644, NULL, &proc_dointvec_jiffies},
+         {NET_IPV4_VS_CACHE_BYPASS, "cache_bypass",
+          &sysctl_ip_vs_cache_bypass, sizeof(int), 0644, NULL,
+          &proc_dointvec},
          {0}},
         {{NET_IPV4_VS, "vs", NULL, 0, 0555, ipv4_vs_table.vs_vars},
          {0}},
@@ -1332,7 +1337,7 @@ static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
                 for (e=l->next; e!=l; e=e->next) {
                         svc = list_entry(e, struct ip_vs_service, s_list);
                         size = sprintf(buf+len, "%s  %08X:%04X %s",
-                                       vs_proto_name(svc->protocol),
+                                       ip_vs_proto_name(svc->protocol),
                                        ntohl(svc->addr),
                                        ntohs(svc->port),
                                        svc->scheduler->name);
@@ -1444,7 +1449,7 @@ static int ip_vs_get_info(char *buf, char **start, off_t offset, int length)
  */
 static int ip_vs_set_timeouts(struct ip_vs_rule_user *u)
 {
-        IP_VS_DBG(1, "Setting timeout tcp:%d tcpfin:%d udp:%d\n",
+        IP_VS_DBG(2, "Setting timeout tcp:%d tcpfin:%d udp:%d\n",
                   u->tcp_timeout/HZ,
                   u->tcp_fin_timeout/HZ,
                   u->udp_timeout/HZ);
@@ -1522,13 +1527,13 @@ do_ip_vs_set_ctl(struct sock *sk, int cmd, void *user, unsigned int len)
 	 * Lookup the exact service by (protocol, vaddr, vport) or fwmark
 	 */
 	if (urule->vfwmark == 0)
-		svc = __ip_vs_lookup_service(urule->protocol,
-					     urule->vaddr, urule->vport);
+		svc = __ip_vs_service_get(urule->protocol,
+                                          urule->vaddr, urule->vport);
 	else
-		svc = __ip_vs_lookup_svc_fwm(urule->vfwmark);
+		svc = __ip_vs_svc_fwm_get(urule->vfwmark);
 
 	if (svc)
-                atomic_dec(&svc->refcnt);
+                ip_vs_service_put(svc);
 
 	switch (cmd) {
 	case IP_VS_SO_SET_ADD:
@@ -1614,7 +1619,7 @@ int __init ip_vs_control_init(void)
         int ret;
         int idx;
 
-        EnterFunction(4);
+        EnterFunction(2);
         
  	ret = nf_register_sockopt(&ip_vs_sockopts);
 	if (ret) {
@@ -1638,15 +1643,15 @@ int __init ip_vs_control_init(void)
                 INIT_LIST_HEAD(&ip_vs_rtable[idx]);
         }
 
-        LeaveFunction(4);
+        LeaveFunction(2);
         return 0;
 }
 
 void __exit ip_vs_control_cleanup(void)
 {
-        EnterFunction(4);
+        EnterFunction(2);
         unregister_sysctl_table(ipv4_vs_table.sysctl_header);
         proc_net_remove("ip_vs");
         nf_unregister_sockopt(&ip_vs_sockopts);
-        LeaveFunction(4);
+        LeaveFunction(2);
 }
